
import pandas as pd


def process(df: pd.DataFrame, mod_by: int = 2):
    result = pd.DataFrame(df["TextField"].str.len() % mod_by)

    result.columns = ["Result"]
    result.insert(0, "Id", df["Id"])


    return result



import re
import asyncio
import os
import types
import typing

import pandas as pd

from . import errors

class NameValidator:
    __valid_name_regex = re.compile(r'^[A-Z][\w\d]*$')

    @classmethod
    def raise_if_invalid(cls, name: str):
        if not cls.__valid_name_regex.match(name):
            raise ValueError("Name is not valid: '{}'!".format(name))

        pass
    


class ServiceContext:
    def get_parameter_value(self, name: str, required: bool = True, default: str = None) -> str:
        raise NotImplementedError()

class ProcessContext:
    def get_parameter_value(self, name: str, required: bool = True, default: str = None) -> str:
        raise NotImplementedError()
    
    def get_input_dataframe(self, name: str, required: bool = True) -> types.CoroutineType:
        raise NotImplementedError()

    def set_output_dataframe(self, name: str, df: pd.DataFrame) -> types.CoroutineType:
        raise NotImplementedError()

class CollectingProcessContext(ProcessContext):
    def __init__(self):
        super()
        self.__output_dataframes = dict()

    async def set_output_dataframe(self, name: str, df: pd.DataFrame):
        NameValidator.raise_if_invalid(name)

        self.__output_dataframes[name] = df
    
    def get_output_dataframe(self, name: str):
        NameValidator.raise_if_invalid(name)

        return self.__output_dataframes.get(name)

    def output_dataframes(self):
        return self.__output_dataframes.items()


class EnvironmentVariableServiceContext(ServiceContext):
    def __init__(self, prefix: str, default_values: dict = None):
        self.__prefix = prefix
        self.__default_values = default_values or dict()
    
    def get_parameter_value(self, name: str, required: bool = True, default: str = None) -> str:
        NameValidator.raise_if_invalid(name)

        ev = os.environ.get(self.__prefix + name)
        if ev:
            return ev

        if name in self.__default_values:
            return self.__default_values[name]

        if required and default is None:
            raise errors.MissingParameterError(name)

        return default



import argparse
import os

from .. import server
from . import local


#https://stackoverflow.com/a/42355279/1270504
class _StoreDictKeyPair(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        my_dict = {}
        for kv in values.split(","):
            k,v = kv.split("=")
            my_dict[k] = v
        setattr(namespace, self.dest, my_dict)

parser = argparse.ArgumentParser(description='Locally debug.', prog = "mlservicewrapper.core.debug")
parser.add_argument(
    '--config', help='Path to service configuration file', required=True)
parser.add_argument('--input-dir', dest='input_dir',
                    help='Path to input directory', required=True)
parser.add_argument('--output-dir', dest='output_dir',
                    help='Path to input directory')
parser.add_argument("--load-params", dest="load_params", action=_StoreDictKeyPair, metavar="KEY1=VAL1,KEY2=VAL2...")
parser.add_argument("--run-params", dest="runtime_parameters", action=_StoreDictKeyPair, metavar="KEY1=VAL1,KEY2=VAL2...")

parser.add_argument('--split-dataset-for-perf', dest='split_dataset_name',
                    help='Input dataset to split for performance evaluation.')

parser.add_argument('--assess-accuracy', dest='assess_accuracy', action=_StoreDictKeyPair, metavar="INPUT_DATASET_1.FIELD=OUTPUT_DATASET_1.FIELD,INPUT_DATASET_2.FIELD=OUTPUT_DATASET_2.FIELD,...",
                    help='Perform accuracy assessment against the given fields.')

args = parser.parse_args()

service, config_params = server.get_service_instance(args.config)

load_params = dict()

if config_params:
    load_params.update(config_params)

if args.load_params:
    load_params.update(args.load_params)
    
local.run(
    service,
    args.input_dir,
    split_dataset_name=args.split_dataset_name,
    load_parameters=args.load_params,
    runtime_parameters=args.runtime_parameters,
    output_file_directory=args.output_dir,
    assess_accuracy=args.assess_accuracy
)



import importlib.util
import json
import os
import sys
from types import SimpleNamespace

from .contexts import EnvironmentVariableServiceContext
from .services import Service


def get_service_instance(config_path: str = None) -> (Service, dict):
    if not config_path:
        config_path = os.environ.get("SERVICE_CONFIG_PATH", "./service/config.json")

    with open(config_path, "r") as config_file:
        config = json.loads(config_file.read())

    service_script_path = config["modulePath"]

    if service_script_path is None:
        raise ValueError("The modulePath couldn't be determined!")

    config_directory_path = os.path.dirname(config_path)

    service_script_path = os.path.realpath(
        os.path.join(config_directory_path, service_script_path))

    print("Loading from script {}".format(service_script_path))

    service_script_dirname = os.path.dirname(service_script_path)
    service_script_basename = os.path.basename(service_script_path)

    os.sys.path.insert(0, service_script_dirname)

    service_script_module_name = os.path.splitext(service_script_basename)[0]

    print("Importing module {} from {}...".format(service_script_module_name, service_script_dirname))

    service_module = importlib.import_module(service_script_module_name)

    print("Imported module")

    if "className" in config:
        service_class_name = config["className"]
        service_type = getattr(service_module, service_class_name)

        print("Identified service type: {}".format(str(service_type)))

        service = service_type()
    else:
        service_instance_name = config["serviceInstanceName"]
        service = getattr(service_module, service_instance_name)

    print("Got service: {}".format(service))

    return (service, config.get("parameters", dict()))



import asyncio
import os
import time

from pandas import DataFrame

import dal
import mlservicewrapper
import mlservicewrapper.core.contexts
import mlservicewrapper.core.services
import mlservicewrapper.core.errors


class SampleService(mlservicewrapper.core.services.Service):
    async def load(self, ctx: mlservicewrapper.core.contexts.ServiceContext):
        self.__mod_by = int(ctx.get_parameter_value("ModBy", default="2"))

    async def process(self, ctx: mlservicewrapper.core.contexts.ProcessContext):
        input_data = await ctx.get_input_dataframe("Data")

        if "TextField" not in input_data.columns:
            raise mlservicewrapper.core.errors.MissingDatasetFieldError("Data", "TextField")

        result = dal.process(input_data, self.__mod_by)

        await ctx.set_output_dataframe("Results", result)



import os

import pandas as pd

import mlservicewrapper
import mlservicewrapper.core.contexts
import mlservicewrapper.core.services
import mlservicewrapper.core.errors

import fasttext

import urllib3
import shutil

import json

class FastTextServiceBase(mlservicewrapper.core.services.Service):
    async def load(self, ctx: mlservicewrapper.core.contexts.ServiceContext):
        model_path = self.get_model_path(ctx)
        
        if not os.path.exists(model_path):
            url = self.get_model_url(ctx)

            http = urllib3.PoolManager()
            r = http.request('GET', url, preload_content=False)

            with open(model_path, 'wb') as out:
                shutil.copyfileobj(r, out)

            r.release_conn()

        self.model = fasttext.load_model(model_path)

    def get_model_path(self, ctx: mlservicewrapper.core.contexts.ServiceContext):
        return ctx.get_parameter_value("ModelPath", required=True)
    
    def get_model_url(self, ctx: mlservicewrapper.core.contexts.ServiceContext):
        return ctx.get_parameter_value("ModelUrl", required=True)

class FastTextVectorizerService(FastTextServiceBase):

    async def process(self, ctx: mlservicewrapper.core.contexts.ProcessContext):
        input_data = await ctx.get_input_dataframe("Data")

        if "Text" not in input_data.columns:
            raise mlservicewrapper.core.errors.MissingDatasetFieldError("Data", "Text")

        input_data["Vector"] = input_data["Text"].str.replace("\n", "").apply(lambda x: json.dumps(self.model.get_sentence_vector(x).tolist()))
        input_data.drop(["Text"], inplace=True, axis="columns")

        await ctx.set_output_dataframe("Results", input_data)

class LanguageDetectionService(FastTextServiceBase):

    def predict(self, text: str):
        t = self.model.predict(text)

        label: str = t[0][0]
        score: float = t[1][0]

        label = label.replace("__label__", "")

        return (label, score)

    async def process(self, ctx: mlservicewrapper.core.contexts.ProcessContext):
        input_data = await ctx.get_input_dataframe("Data")
        id_field = ctx.get_parameter_value("IdFieldName", required=False)

        if "Text" not in input_data.columns:
            raise mlservicewrapper.core.errors.MissingDatasetFieldError("Data", "Text")

        if id_field is not None and id_field not in input_data.columns:
            raise mlservicewrapper.core.errors.MissingDatasetFieldError("Data", id_field, "The field specified by parameter 'IdFieldName' was not present in the dataset.")

        results = pd.DataFrame(input_data["Text"].str.replace("\n", "")).apply(lambda x: self.predict(x[0]), result_type="expand", axis="columns")
        results.columns = ["Label", "Score"]
        results.set_index(input_data.index)

        if id_field is not None:
            results.insert(0, id_field, input_data[id_field])

        await ctx.set_output_dataframe("Results", results)

    def get_model_url(self, ctx: mlservicewrapper.core.contexts.ServiceContext):
        return "https://dl.fbaipublicfiles.com/fasttext/supervised-models/lid.176.bin"



import math
import statistics
import asyncio
import json
import os
import re
import typing
import time

import pandas as pd

from .. import contexts, errors, services

def _print_ascii_histogram(seq: typing.List[float]) -> None:
    """A horizontal frequency-table/histogram plot."""

    hist = {}

    _min = min(seq)
    _max = max(seq)
    _len = len(seq)

    buckets = 10
    step = (_max - _min) / (buckets - 1)

    for i in seq:
        e = _min + (math.floor((i - _min) / step) * step)

        hist[e] = hist.get(e, 0) + 1

    for i in range(buckets):
        e = _min + (i * step)

        ct = hist.get(e, 0)

        pct = ct / _len

        w = math.floor(40 * pct)

        if ct > 0:
            w = max(w, 1)

        print('{0:5f}s {1}'.format(e, '+' * w))

class _LocalLoadContext(contexts.ServiceContext):
    def __init__(self, parameters: dict = None):
        self.__parameters = parameters or dict()

        print("Loaded service parameters:")
        print(self.__parameters)
        print()
            
    def get_parameter_value(self, name: str, required: bool = True, default: str = None) -> str:
        contexts.NameValidator.raise_if_invalid(name)
        
        if name in self.__parameters:
            return self.__parameters[name]
        
        if required and default is None:
            raise errors.MissingParameterError(name)

        print("Could not find optional parameter {}".format(name))

        return default

def get_input_path(dir_path: str, name: str):
    name_regex = re.escape(name) + r"\.\w+"
    
    file_path: str = None
    for f in os.scandir(dir_path):
        if not re.match(name_regex, f.name):
            continue

        if file_path:
            raise ValueError("Multiple files matched input dataset {}".format(name))

        file_path = f.path

    return file_path

def get_input_dataframe(dir_path: str, name: str):
    contexts.NameValidator.raise_if_invalid(name)
    
    file_path = get_input_path(dir_path, name)

    if file_path:
        return pd.read_csv(file_path)

    return None

class _LocalRunContext(contexts.CollectingProcessContext):
    def __init__(self, input_files_dir: str, output_files_dir: str = None, parameters: dict = None):
        super().__init__()
        self.__parameters = parameters or dict()

        self.__input_files_dir = input_files_dir
        self.__output_files_dir = output_files_dir

    def get_parameter_value(self, name: str, required: bool = True, default: str = None) -> str:
        contexts.NameValidator.raise_if_invalid(name)
        
        if name in self.__parameters:
            return self.__parameters[name]
        
        if required and default is None:
            raise errors.MissingParameterError(name)
            
        print("Could not find optional parameter {}".format(name))

        return default

    async def get_input_dataframe(self, name: str, required: bool = True):
        contexts.NameValidator.raise_if_invalid(name)
        
        df = get_input_dataframe(self.__input_files_dir, name)
        
        if required and df is None:
            raise errors.MissingDatasetError(name)

        return df

    async def set_output_dataframe(self, name: str, df: pd.DataFrame):
        contexts.NameValidator.raise_if_invalid(name)
        
        await super().set_output_dataframe(name, df)
        
        # print("Got results for {}".format(name))
        # print(df)
        # print()

        if self.__output_files_dir:
            os.makedirs(self.__output_files_dir, exist_ok=True)

            df.to_csv(os.path.join(self.__output_files_dir, name + ".csv"), index=False)

class _LocalDataFrameRunContext(contexts.ProcessContext):
    def __init__(self, df: pd.DataFrame, name: str, base_ctx: contexts.ProcessContext):
        self.__base_ctx = base_ctx
        self.__name = name
        self.__df = df

    def get_parameter_value(self, name: str, required: bool = True, default: str = None) -> str:
        return self.__base_ctx.get_parameter_value(name, required, default)

    def set_output_dataframe(self, name: str, df: pd.DataFrame):
        return self.__base_ctx.set_output_dataframe(name, df)

    async def get_input_dataframe(self, name: str, required: bool = True):
        contexts.NameValidator.raise_if_invalid(name)
        
        if name == self.__name:
            return self.__df

        return await self.__base_ctx.get_input_dataframe(name, required)

    def output_dataframes(self):
        return self.__base_ctx.output_dataframes()

async def _perform_accuracy_assessment(ctx: contexts.CollectingProcessContext, specs: dict):

    for k, v in specs.items():
        i = k.split(".")
        o = v.split(".")

        input_df = await ctx.get_input_dataframe(i[0], required=True)
        output_df = ctx.get_output_dataframe(o[0])

        input_field = input_df[i[1]]
        input_field.name = "Expected"

        output_field = output_df[o[1]]
        output_field.name = "Actual"

        joined = output_field.to_frame().join(input_field, how="inner")

        joined["Result"] = joined["Actual"] == joined["Expected"]

        count_total = len(joined.index)
        count_correct = joined["Result"].values.sum()

        print("Accuracy ({} to {}): {} of {} ({})".format(k, v, count_correct, count_total, count_correct / count_total))


async def run_async(service: typing.Union[services.Service, typing.Callable], input_file_directory: str, output_file_directory: str = None, split_dataset_name: str = None, load_parameters: dict = None, runtime_parameters: dict = None, assess_accuracy: dict = None):
    if callable(service):
        service = service()
        initialized_service = True
    else:
        initialized_service = False
    
    load_context = _LocalLoadContext(load_parameters)

    if hasattr(service, 'load'):
        print("Loading...")
        s = time.perf_counter()
        await service.load(load_context)
        e = time.perf_counter()

        load_time = e - s
    else:
        load_time = 0
    
    print("Running...")

    run_context = _LocalRunContext(input_file_directory, output_file_directory, runtime_parameters)

    times = list()

    if split_dataset_name:
        split_dataset_path = get_input_path(input_file_directory, split_dataset_name)
        
        df = pd.read_csv(split_dataset_path)

        for r in df.itertuples(index=False):
            rdf = pd.DataFrame([r])
        
            row_run_context = _LocalDataFrameRunContext(rdf, split_dataset_name, run_context)

            s = time.perf_counter()
            await service.process(row_run_context)
            e = time.perf_counter()

            times.append(e - s)
    else:

        s = time.perf_counter()
        await service.process(run_context)
        e = time.perf_counter()

        times.append(e - s)

    print("Load time: {}s".format(load_time))
    if len(times) == 1:
        print("Process time: {}s".format(times[0]))
    else:
        print()
        print("Count: {}".format(len(times)))
        print("Min process time: {}s".format(min(times)))
        print("Mean process time: {}s".format(statistics.mean(times)))
        print("Median process time: {}s".format(statistics.median(times)))
        print("Max process time: {}s".format(max(times)))

        _print_ascii_histogram(times)

    if initialized_service and hasattr(service, 'dispose'):
        service.dispose()
    
    result = dict(run_context.output_dataframes())

    if assess_accuracy is not None:
        await _perform_accuracy_assessment(run_context, assess_accuracy)

    return result

def run(service: typing.Union[services.Service, typing.Callable], input_file_directory: str, output_file_directory: str = None, split_dataset_name: str = None, load_parameters: dict = None, runtime_parameters: dict = None, assess_accuracy: dict = None):

    loop = asyncio.get_event_loop()

    return loop.run_until_complete(run_async(service, input_file_directory, output_file_directory, split_dataset_name, load_parameters, runtime_parameters, assess_accuracy))



import os
from setuptools import find_namespace_packages, setup, find_packages

if os.environ.get("USE_SCM_VERSION"):
   use_scm_version = True
   version = None
else:
   use_scm_version = False
   version = "0.0.0"

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
   name='mlservicewrapper-core',
   version = version,
   use_scm_version = use_scm_version,
   description='Configure a Python service for repeated execution',
   author='Matthew Haugen',
   author_email='mhaugen@haugenapplications.com',

   url="https://github.com/MaJaHa95/ml-service-wrapper",
   long_description=long_description,
   long_description_content_type="text/markdown",

   package_dir={"": "src"},
   packages=find_namespace_packages("src", include=['mlservicewrapper.*']),

   install_requires=[
      "pandas"
   ],
   #namespace_packages=['mlservicewrapper.core.errors'],
   setup_requires=['setuptools_scm'],
   zip_safe=False,
   python_requires='>=3.6'
)
